/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:10:54 1991 by burrows        */
#include <stdio.h>
#include "indexserver.h"

struct msgset *notmsgset (a, b)
struct msgset *a, *b;
{
	struct msgset *result;
	register long *x, *y;
	register int i;

	if (a == 0)
		return (0);

	result = newmsgset (a->nbitset);

	bcopy ((char *) a->bitset, (char *) result->bitset,
		a->nbitset * sizeof (long));
	
	if (b == 0)
		return (result);

	x = result->bitset;
	y = b->bitset;
	i = MIN (result->nbitset, b->nbitset);
	i++;
	while (--i != 0)
		*x++ &= ~*y++;

	return (result);
}

struct msgset *andmsgset (a, b)
register struct msgset *a, *b;
{
	struct msgset *result;
	register long *x, *y;
	register int i;

	if (a == 0 || b == 0)
		return (0);

	i = MIN (a->nbitset, b->nbitset);
	result = newmsgset (i);

	bcopy ((char *) a->bitset, (char *) result->bitset, i * sizeof (long));
	
	x = result->bitset;
	y = b->bitset;
	i++;
	while (--i != 0)
		*x++ &= *y++;

	return (result);
}

struct msgset *ormsgset (a, b)
register struct msgset *a, *b;
{
	struct msgset *result;
	register long *x, *y;
	register int i;

	if (a == 0) {
		a = b;
		b = 0;
	}

	if (a == 0)
		return (0);
	
	if (b == 0)
		i = a->nbitset;
	else
		i = MAX (a->nbitset, b->nbitset);

	result = newmsgset (i);

	bcopy ((char *) a->bitset, (char *) result->bitset, i * sizeof (long));

	if (b == 0)
		return (result);
	
	x = result->bitset;
	y = b->bitset;
	i++;
	while (--i != 0)
		*x++ |= *y++;

	return (result);
}

struct msgset *into_notmsgset (a, b)
register struct msgset *a, *b;
{
	register long *x, *y;
	register int i;

	if (a == 0 || b == 0)
		return (a);

	i = MIN (a->nbitset, b->nbitset);
	x = a->bitset;
	y = b->bitset;

	i++;
	while (--i != 0)
		*x++ &= ~*y++;

	return (a);
}

struct msgset *into_andmsgset (a, b)
register struct msgset *a, *b;
{
	register long *x, *y;
	register int i;

	if (a == 0)
		return (a);
	if (b == 0) {
		freemsgset (a);
		return (0);
	}

	i = MIN (a->nbitset, b->nbitset);
	x = a->bitset;
	y = b->bitset;

	i++;
	while (--i != 0)
		*x++ &= *y++;

	if (a->nbitset > b->nbitset) {
		i = a->nbitset - b->nbitset;
		i++;
		while (--i != 0)
			*x++ = 0;
	}

	return (a);
}

struct msgset *into_ormsgset (a, b)
register struct msgset *a, *b;
{
	register long *x, *y;
	register int i;

	if (b == 0)
		return (a);

	if (a == 0)
		return (ormsgset (b, (struct msgset *) 0));

	if (a->nbitset < b->nbitset)
		a = extendmsgset (a, b->nbitset);
	
	i = MIN (a->nbitset, b->nbitset);
	x = a->bitset;
	y = b->bitset;

	i++;
	while (--i != 0)
		*x++ |= *y++;

	return (a);
}

struct msgset *extendmsgset (msgset, nbitset)
register struct msgset *msgset;
int nbitset;
{
	int old;

	if (msgset == 0) {
		msgset = newmsgset (nbitset);
		old = 0;
	} else {
		old = msgset->nbitset;

		if (old >= nbitset)
			return (msgset);

		msgset->nbitset = nbitset;
		msgset->bitset = (long *) realloc ((char *) msgset->bitset,
			(unsigned) (sizeof (long) * msgset->nbitset));

		if (msgset->bitset == 0) {
			perror ("indexserver: realloc newmsgset bitset");
			exit (2);
		}
	}

	bzero ((char *) &msgset->bitset[old], (int) ((msgset->nbitset - old) *
			    sizeof (long)));

	return (msgset);
}

static int msgsetcount = 0;

struct msgset *newmsgset (nbitset)
int nbitset;
{
	register struct msgset *new;
	long *bitset;

	if (nbitset < 1)
		nbitset = 1;

	if ((new = (struct msgset *) malloc (sizeof (struct msgset))) == 0) {
		perror ("indexserver: malloc newmsgset");
		exit (2);
	}
	if ((bitset = (long *)malloc ((unsigned) (nbitset*sizeof (long))))==0) {
		perror ("indexserver: malloc newmsgset bitset");
		exit (2);
	}

	new->nbitset = nbitset;
	new->bitset = bitset;
/* (void) fprintf (stderr, "msgset %d\n", ++msgsetcount); */

	return (new);
}

void freemsgset (m)
register struct msgset *m;
{
	if (m != 0) {
		free ((char *) m->bitset);
		free ((char *) m);
/* (void) fprintf (stderr, "msgset %d\n", --msgsetcount); */
	}
}

long countmsgset (msgset)
struct msgset *msgset;
{
	register long *bitset;
	register int i;
	register long count = 0;
	static char set[256];
	static int first = 1;
	register char *rset = set;

	if (first) {
		first = 0;
		for (i = 0; i < 256; i++) {
			register int k = 0;
			register int j = i;
			while (j != 0) {
				if (j & 1)
					k++;
				j >>= 1;
			}
			rset[i] = k;
		}
	}

	if (msgset != 0) {
		register long k;

		i = msgset->nbitset;
		bitset = msgset->bitset;
		i++;
		while (--i != 0) {
			if ((k = *bitset++) != 0) {
#if LONGBITS >= 8
				count += rset[k & 0xff];
#endif
#if LONGBITS >= 16
				count += rset[(k >> 8) & 0xff];
#endif
#if LONGBITS >= 24
				count += rset[(k >> 16) & 0xff];
#endif
#if LONGBITS >= 32
				count += rset[(k >> 24) & 0xff];
#endif
#if LONGBITS >= 40
				count += rset[(k >> 32) & 0xff];
#endif
#if LONGBITS >= 48
				count += rset[(k >> 40) & 0xff];
#endif
#if LONGBITS >= 56
				count += rset[(k >> 48) & 0xff];
#endif
#if LONGBITS >= 64
				count += rset[(k >> 56) & 0xff];
#endif
			}
		}
	}

	return (count);
}

int applymsgset (func, index2, groups, msgset)
int (*func) ();
struct index2 *index2;
struct groups *groups;
struct msgset *msgset;
{
	register long *bitset;
	register int i;

	if (msgset != 0) {
		register long k;
		register long j;

		i = msgset->nbitset;
		bitset = msgset->bitset;
		i++;
		while (--i != 0) {
			if ((j = *bitset++) != 0) {
				k = (msgset->nbitset - i) * LONGBITS;
				if (j & 1)
					if ((*func) (index2, groups, k) != 0)
						return (-1);
				j >>= 1;
				j &= ~(1 << (LONGBITS-1));
				k++;
				while (j != 0) {
					if (j & 1)
						if ((*func)
						    (index2, groups, k) != 0)
							return (-1);
					j >>= 1;
					k++;
				}
			}
		}
	}

	return (0);
}
