/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:49:42 1991 by burrows        */
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <strings.h>
#include "indexserver.h"

static int readconcordance ();

void freeindex (index)
register struct index *index;
{
	freebinfile (&index->summary.messages);
	freebinfile (&index->summary.dates);
	freetextfile (&index->summary.summary);
	if (index->summary.fields != 0) {
		free (index->summary.fields);
		index->summary.fields = 0;
	}
	freebinfile (&index->concord.data);
	if (index->concord.fd > 0)	/* so we don't close stdin by mistake */
		(void) close (index->concord.fd);
	index->concord.fd = -1;
	freetextfile (&index->concord.subindex);
}

int readindex (index, pos, messagesfile, concordancefile, summaryfile,datesfile)
struct index *index;
long pos;
char *messagesfile;
char *concordancefile;
char *summaryfile;
char *datesfile;
{
	register long *x;
	register int i, k;
	register char *p, *f, *start;

	if (readbinfile (messagesfile, &index->summary.messages, pos) < 0)
		return (-1);

	if (readbinfile (datesfile, &index->summary.dates, 0L) < 0)
		return (-1);

	if (readtextfile (summaryfile, &index->summary.summary, 0L) < 0)
		return (-1);

	index->summary.fields = malloc ((unsigned)
		(index->summary.summary.nlines * NFIELDS));
	if (index->summary.fields == 0) {
		perror ("indexserver: readindex: fields: malloc");
		return (-1);
	}
	bzero (index->summary.fields,
		(int) (index->summary.summary.nlines * NFIELDS));

	i = index->summary.summary.nlines;
	f = index->summary.fields;
	x = index->summary.summary.lines;
	i++;
	while (--i != 0) {
		p = start = *x++ + index->summary.summary.text.data;
		for (k = 0; k < NFIELDS; k++, f++) {
			while (*p != '\0' && *p != '\t')
				p++;
			if (*p != '\t') {
				    (void) fprintf (stderr,
					"newsindex: malformed summary file %s, line %d\n",
					summaryfile,
					index->summary.summary.nlines - i + 1);
abort ();
					return (-1);
			}
			*p++ = '\0';
			if (p - start < 256)
				*f = p - start;
		}
	}

	return (readconcordance (concordancefile, index));
}

static int readconcordance (file, index)
char *file;
struct index *index;
{
	struct stat info;
	char ifile[1025];
	register int i, n;
	register long *l;
	register char *data, *p;

	if (stat (file, &info) < 0) {
		perror (file);
		return (-1);
	}

	if (info.st_size > 1024*1024) {
		index->concord.fd = open (file, 0);
		index->concord.data.ndata = info.st_size;
		if (index->concord.data.data != 0) {
			(void) fprintf (stderr,
			   "indexserver: readconcordance: data non-zero\n");
			abort ();
		}
		if (index->concord.fd < 0) {
			perror (file);
			return (-1);
		}
	} else if (readbinfile (file, &index->concord.data, 0L) < 0)
		return (-1);
	
	(void) strcpy (ifile, file);
	(void) strcat (ifile, ".i");
	if (readtextfile (ifile, &index->concord.subindex, 0L) < 0)
		return (-1);
	
	data = index->concord.subindex.text.data;
	l = index->concord.subindex.lines;
	for (i = 0, n = index->concord.subindex.nlines; i < n; i++) {
		p = rindex (&data[*l++], ' ');
		if (p == 0) {
			(void) fprintf (stderr,
				"indexserver: %s: bad format: line %d\n",
				ifile, i+1);
			return (-1);
		}
		*p = '\0';
	}

	return (0);
}

void freegroups (groups)
register struct groups *groups;
{
	freetextfile (&groups->text);
	if (groups->group != 0) {
		free ((char *) groups->group);
		groups->group = 0;
	}
}

int readgroups (groups, groupsfile)
register struct groups *groups;
char *groupsfile;
{
	register int old,  maxgroup = 0;
	register int i;
	char *data;
	long *lines;
	char buf[500];
	int one, two, num, lastmessage, firstmessage;
	char mode[256];
	register char *p;

	if (readtextfile (groupsfile, &groups->text, 0L) < 0)
		return (-1);

	old = maxgroup;
	maxgroup = groups->text.nlines+10;
	groups->group = (long *) malloc ((unsigned) (maxgroup * sizeof (long)));
	if (groups->group == 0) {
		perror ("indexserver: readgroups: malloc");
		return (-1);
	}
	while (old < maxgroup)
		groups->group[old++] = -1;

	data = groups->text.text.data;
	lines = groups->text.lines;
	for (i = 0; i < groups->text.nlines; i++) {
		int count = sscanf (data+lines[i], "%63s %d %d %s %d %d %d",
			buf, &one, &two, mode,
			&num, &lastmessage, &firstmessage);
		if (count != 7) {
			(void) fprintf (stderr,
				"indexserver: bad group file: %s\n",
				data+lines[i]);
			return (-1);
		}
		for (p = buf; *p != '\0'; p++)
			if (*p == '.')
				*p = '/';
		(void) strcpy (data+lines[i], buf);
		if (num >= maxgroup) {
			old = maxgroup;
			maxgroup = num + 10;
			groups->group = (long *) realloc ((char *)groups->group,
				(unsigned) (maxgroup * sizeof (long)));
			if (groups->group == 0) {
				perror ("indexserver: readgroups: realloc");
				return (-1);
			}
			while (old < maxgroup)
				groups->group[old++] = -1;
		}
		groups->group[num] = groups->text.lines[i];
	}
	groups->ngroup = maxgroup;

	return (0);
}

int readbinfile (file, binfile, pos)
char *file;
struct binfile *binfile;
long pos;
{
	int fd = open (file, 0);
	struct stat info;
	int got;
	int size;
	char *data;
	int nread;

	if (fd < 0) {
		perror (file);
		return (-1);
	}

	if (fstat (fd, &info) < 0) {
		perror (file);
		(void) close (fd);
		return (-1);
	}
	if (pos != 0L && lseek (fd, pos, 0) == -1L) {
		perror (file);
		(void) close (fd);
		return (-1);
	}
	size = info.st_size - pos;
	if (size < 0)
		size = 0;
	size += 1024;

	data = malloc ((unsigned) size + 1);

	if (data == 0) {
		perror ("indexserver: malloc binfile data");
		(void) close (fd);
abort ();
		return (-1);
	}

	got = 0;
	while ((nread = read (fd, data+got, size - got)) > 0) {
		got += nread;
		if (got > size) {
			size = got + 8192;
			data = realloc (data, (unsigned) size + 1);
			if (data == 0) {
				perror ("indexserver: realloc binfile data");
				(void) close (fd);
				return (-1);
			}
		}
	}

	if (nread < 0) {
		perror (file);
		(void) close (fd);
		free (data);
		return (-1);
	}

	(void) close (fd);

	binfile->data = data;
	binfile->ndata = got;

	return (0);
}

int readtextfile (file, textfile, pos)
char *file;
struct textfile *textfile;
long pos;
{
	struct binfile *binfile = &textfile->text;
	register int i;
	register char *p, *e;
	register int maxlines;
	register long *lines;

	if (readbinfile (file, binfile, pos) < 0)
		return (-1);

	maxlines = (binfile->ndata / 26) + 10;
	lines = (long *) malloc ((unsigned) (maxlines * sizeof (long)));
	if (lines == 0) {
		perror ("indexserver: malloc textfile lines");
		freebinfile (binfile);
		free ((char *) textfile);
		return (-1);
	}

	binfile->data[binfile->ndata] = '\n';
	for (i = 0, p = binfile->data, e = p+binfile->ndata; p < e; ) {
		lines[i++] = p - binfile->data;
		while (*p++ != '\n' &&  *p++ != '\n' &&
		       *p++ != '\n' &&  *p++ != '\n')
			;
		p[-1] = '\0';
		if (i >= maxlines) {
			maxlines += 1024;
			lines = (long *) realloc ((char *) lines,
				(unsigned) maxlines * sizeof (long));
			if (lines == 0) {
				perror ("indexserver: realloc textfile lines");
				freebinfile (binfile);
				free ((char *) textfile);
				return (-1);
			}
		}
	}
	textfile->nlines = i;
	textfile->lines = lines;

	return (0);
}

void freebinfile (binfile)
struct binfile *binfile;
{
	if (binfile->data != 0) {
		free (binfile->data);
		binfile->data = 0;
	}
}

void freetextfile (textfile)
struct textfile *textfile;
{
	freebinfile (&textfile->text);
	if (textfile->lines != 0) {
		free ((char *) textfile->lines);
		textfile->lines = 0;
	}
}
