/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Wed Feb 20 17:08:00 PST 1991 by burrows    */
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include "indexserver.h"

struct shellmeta {
	int type;
	struct shellmeta *next;
	char table[256];
};

char *compileshellmeta (s, ignorecase)
char *s;
int ignorecase;
{
	struct shellmeta *front = 0;
	register struct shellmeta *current = 0;
	register struct shellmeta *new;
	register int i;

	while (*s != '\0') {
		new = (struct shellmeta *) malloc (sizeof (struct shellmeta));
		if (new == 0) {
			freeshellmeta ((char *) front);
			return (0);
		}
		new->next = 0;
		if (current != 0)
			current->next = new;
		if (front == 0)
			front = new;
		current = new;

		current->type = *s & 0xff;
		if (*s == '*') {
			while (*s == '*')
				s++;
		} else if (*s == '?') {
			for (i = 0; i < 256; i++)
				current->table[i] = 1;
			s++;
		} else if (*s == '[' && s[1] != '\0' && s[2] != '\0' &&
			   index (s+(s[1]=='!'?3:2), ']') != 0) {
                        register int not = 0, j;

			bzero (current->table, 256);
                        if (*++s == '!') {
                                not = 1;
                                s++;
                        }

                        do {
				if (ignorecase && isalpha (*s & 0xff)) {
					current->table[tolower (*s & 0xff)] = 1;
					current->table[toupper (*s & 0xff)] = 1;
					s++;
				} else
					current->table[*s++ & 0xff] = 1;
                                if (*s == '-' && s[1] != ']') {
					i = (*(s-1) & 0xff);
					j = (*(s+1) & 0xff);
					if (ignorecase) {
						while (i <= j) {
							if (isalpha (i)) {
								current->table[toupper (i)] = 1;
								current->table[tolower (i)] = 1;
							} else
								current->table[i] = 1;
							i++;
						}
					} else
						while (i <= j)
							current->table[i++] = 1;
                                        s += 2;
                                }
                        } while (*s != ']');
			s++;

                        if (not) {
				for (i = 0; i < 256; i++)
					current->table[i] = !current->table[i];
			}
		} else {
			bzero (current->table, 256);
			if (ignorecase && isalpha (*s & 0xff)) {
				current->type = '[';
				current->table[toupper (*s & 0xff)] = 1;
				current->table[tolower (*s & 0xff)] = 1;
			} else
				current->table[*s & 0xff] = 1;
			s++;
		}
		current->table[0] = 0;
	}

	return ((char *) front);
}

int findstartshellmeta (pat, s, stab, etab)
char *pat;
register char *s;
register char *stab;
char *etab;
{
	register int c;
	register int d;
	register struct shellmeta *current = (struct shellmeta *) pat;

	if (current == 0 || (c = current->type) == '*')
		return (0);

	if (c == '[') {
		if ((d = (*s++ & 0xff)) == 0)
			return (0);

		if (current->table[d] && findshellmeta (pat, s-1, etab))
			return (1);

		for (;;) {
			while ((d = (*s++ & 0xff)) != 0 && !current->table[d] &&
			       (d = (*s++ & 0xff)) != 0 && !current->table[d] &&
			       (d = (*s++ & 0xff)) != 0 && !current->table[d] &&
			       (d = (*s++ & 0xff)) != 0 && !current->table[d])
				;
			if (d == 0)
				return (0);
			if ((stab[s[-2] & 0xff] || stab[s[-1] & 0xff]) &&
			    findshellmeta (pat, s-1, etab))
				return (1);
		}
	} else if (c != '?') {
		if ((d = (*s++ & 0xff)) == 0)
			return (0);

		if (d == c && findshellmeta (pat, s-1, etab))
			return (1);

		for (;;) {
			while ((d = *s++) != 0 && d != c &&
			       (d = *s++) != 0 && d != c &&
			       (d = *s++) != 0 && d != c &&
			       (d = *s++) != 0 && d != c)
				;
			if (d == 0)
				return (0);
			if ((stab[s[-2] & 0xff] || stab[s[-1] & 0xff]) &&
			    findshellmeta (pat, s-1, etab))
				return (1);
		}
	} else {
		if ((d = (*s++ & 0xff)) == 0)
			return (0);

		if (findshellmeta (pat, s-1, etab))
			return (1);

		for (;;) {
			if ((d = (*s++ & 0xff)) == 0)
				return (0);
			if ((stab[s[-2] & 0xff] || stab[s[-1] & 0xff]) &&
			    findshellmeta (pat, s-1, etab))
				return (1);
		}
	}
}

int notfindsimpleshellmeta (pat, s)
char *pat;
register char *s;
{
	register struct shellmeta *current = (struct shellmeta *) pat;
	register struct shellmeta *adv;
	register char *advs;
	register int c;
	register int d;

	while (current != 0) {
		if (current->type == '*') {
			current = current->next;
			if (current == 0) {
				return (0);
			} else {
				if ((c = current->type) == '*') {
					(void) fprintf (stderr,
					  "notfindsimpleshellmeta: two *: %s\n",
						s);
					abort ();
				}

				if (c == '[' || c == '?') {
					while ((d = (*s++ & 0xff)) != 0 &&
						    !current->table[d])
						;
				} else {
					while ((d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c)
						;
				}
				s--;

				while (*s != '\0') {
					adv = current;
					advs = s;
					while (adv != 0 && adv->type != '*' &&
					       adv->table[*advs & 0xff]) {
						adv = adv->next;
						advs++;
					}
					if (adv == 0) {
						if (*advs == '\0')
							return (0);
					} else if (adv->type == '*') {
						if (notfindsimpleshellmeta (
						    (char *) adv, advs) == 0)
							return (0);
					}
					do {
						s++;
					} while (*s != '\0' &&
					       !current->table[*s & 0xff]);
				}
				return (1);
			}
		} else if (current->table[*s++ & 0xff]) {
			current = current->next;
		} else
			return (1);
	}

	return (*s == '\0'?0:1);
}

int findshellmeta (pat, s, tab)
char *pat;
register char *s;
char *tab;
{
	register struct shellmeta *current = (struct shellmeta *) pat;
	register struct shellmeta *adv;
	register char *advs;
	register int c;
	register int d;
	char *origs = s;

	while (current != 0) {
		if (current->type == '*') {
			current = current->next;
			if (current == 0) {
				if (tab[0])
					return (1);
				if (s > origs)
					s--;
				do {
					if (tab[*s & 0xff])
						return (1);
				} while (*s++ != '\0');

				return (0);
			} else {
				if ((c = current->type) == '*') {
					(void) fprintf (stderr,
					  "findeshellmeta: two *: %s\n", s);
					abort ();
				}

				if (c == '[' || c == '?') {
					while ((d = (*s++ & 0xff)) != 0 &&
						    !current->table[d])
						;
				} else {
					while ((d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c &&
					       (d=(*s++ & 0xff))!= 0 && d !=c)
						;
				}
				s--;

				while (*s != '\0') {
					adv = current;
					advs = s;
					while (adv != 0 && adv->type != '*' &&
					       adv->table[*advs & 0xff]) {
						adv = adv->next;
						advs++;
					}
					if (adv == 0) {
						if (tab[*advs & 0xff] ||
						    (advs > origs &&
						     tab[advs[-1] & 0xff]))
							return (1);
					} else if (adv->type == '*') {
						if (findshellmeta (
						    (char *) adv, advs, tab))
							return (1);
					}
					do {
						s++;
					} while (*s != '\0' &&
					       !current->table[*s & 0xff]);
				}
				return (0);
			}
		} else if (current->table[*s++ & 0xff]) {
			current = current->next;
		} else
			return (0);
	}

	return (tab[*s & 0xff] || (s > origs && tab[s[-1] & 0xff]));
}


void freeshellmeta (x)
register char *x;
{
	register struct shellmeta *y = (struct shellmeta *) x;
	register struct shellmeta *save;

	while (y != 0) {
		save = y->next;
		free ((char *) y);
		y = save;
	}
}

#ifdef notdef
printshellmeta (p)
char *p;
{
	struct shellmeta *x = (struct shellmeta *) p;
	int i;

	while (x != 0) {
		if (x->type != '[') {
			putchar (x->type);
		} else {
			putchar ('[');
			for (i = 0; i < 256; i++) {
				if (x->table[i]) {
					if (i >= ' ' && i < 127)
						putchar (i);
					else
						(void) printf ("\\03o", i);
				}
			}
			putchar (']');
		}
		x = x->next;
	}
	putchar ('\n');
}
#endif
